"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var EMPTY_LIT = Object.freeze({
    type: 'Literal',
    value: '',
});
var Parser = /** @class */ (function () {
    function Parser(tokens) {
        this.tokens = tokens;
        this.current = 0;
        this.token = tokens[this.current];
    }
    Parser.prototype.next = function (type, value, strict) {
        if (type === void 0) { type = null; }
        if (value === void 0) { value = null; }
        if (strict === void 0) { strict = true; }
        this.token = this.tokens[++this.current];
        if ((type !== null && this.token.type !== type) ||
            (value !== null && this.token.value !== value)) {
            if (strict) {
                throw new TypeError("Unexpected token: found " + this.token.type + "(" + this.token.value + "), expected " + type + "(" + value + ")");
            }
            else {
                return false;
            }
        }
        return true;
    };
    Parser.prototype.decl = function () {
        var node = {
            type: 'Declaration',
            class: this.token.value,
            name: '',
            parameters: [],
            body: [],
        };
        this.next();
        while (this.token.type !== 'symbol' || this.token.value !== '=') {
            var param = this.param();
            node.parameters.push(param);
        }
        this.next('name');
        node.name = this.token.value;
        if (this.next('symbol', ':', false)) {
            this.next('string');
            node.description = this.literal();
        }
        // @ts-ignore
        if (this.token.type !== 'square' && this.token.value === '[') {
            throw new TypeError("Unexpected token: " + this.token.value);
        }
        this.next();
        while (
        // @ts-ignore
        this.token.type !== 'square' ||
            // @ts-ignore
            (this.token.type === 'square' && this.token.value !== ']')) {
            node.body.push(this.body());
        }
        this.next();
        return node;
    };
    Parser.prototype.body = function () {
        if (this.token.type !== 'name') {
            throw new TypeError("Unexpected token type: " + this.token.type + " " + this.token.value);
        }
        if (this.token.value === 'input' || this.token.value === 'output') {
            return this.connection();
        }
        if (this.token.value.toLowerCase() === 'spawnflags') {
            return this.spawnflags();
        }
        return this.property();
    };
    Parser.prototype.include = function () {
        this.next('string');
        var node = {
            type: 'Include',
            name: this.token.value,
        };
        this.next();
        return node;
    };
    Parser.prototype.mapsize = function () {
        this.next('paren', '(');
        this.next('number');
        var node = {
            type: 'Mapsize',
            x: this.token.value,
            y: 0,
        };
        this.next('symbol', ',');
        this.next('number');
        node.y = this.token.value;
        this.next('paren', ')');
        this.next();
        return node;
    };
    Parser.prototype.materialExclusion = function () {
        this.next('square', '[');
        var node = {
            type: 'MaterialExclusion',
            materials: [],
        };
        while (this.next('string', null, false)) {
            node.materials.push(this.token.value);
        }
        if (this.token.type !== 'square' || this.token.value !== ']') {
            throw new TypeError("Unexpected token " + this.token.type + " " + this.token.value);
        }
        this.next();
        return node;
    };
    Parser.prototype.autoVisGroup = function () {
        this.next('symbol', '=');
        this.next('string');
        var node = {
            type: 'AutoVisGroup',
            description: this.token.value,
            categories: [],
        };
        this.next('square', '[');
        while (this.next('string', null, false)) {
            var category = {
                type: 'Category',
                name: this.token.value,
                groups: [],
            };
            this.next('square', '[');
            while (this.next('string', null, false)) {
                category.groups.push({
                    type: 'VisGroup',
                    name: this.token.value,
                });
            }
            if (this.token.type !== 'square' || this.token.value !== ']') {
                throw new TypeError("Unexpected " + this.token.type + " token " + this.token.value);
            }
            node.categories.push(category);
        }
        this.next();
        return node;
    };
    Parser.prototype.connection = function () {
        var node = {
            type: 'Connection',
            direction: this.token.value,
            name: '',
            args: '',
        };
        this.next('name');
        node.name = this.token.value;
        this.next('paren', '(');
        this.next('name');
        node.args = this.token.value;
        this.next('paren', ')');
        if (this.next('symbol', ':', false)) {
            this.next('string');
            node.description = this.literal();
        }
        return node;
    };
    Parser.prototype.spawnflags = function () {
        var node = {
            type: 'SpawnFlags',
            flags: [],
        };
        this.next('paren', '(');
        this.next('name');
        this.next('paren', ')');
        this.next('symbol', '=');
        this.next('square', '[');
        this.next();
        while (this.token.type !== 'square' ||
            (this.token.type === 'square' && this.token.value !== ']')) {
            var prop = {
                type: 'Flag',
                value: this.literal(),
                title: EMPTY_LIT,
                enabled: EMPTY_LIT,
            };
            if (this.token.type !== 'symbol' || this.token.value !== ':') {
                throw new TypeError("Unexpected token: " + this.token.type + " " + this.token.value);
            }
            this.next();
            prop.title = this.literal();
            if (this.token.type !== 'symbol' || this.token.value !== ':') {
                throw new TypeError("Unexpected token: " + this.token.type + " " + this.token.value);
            }
            this.next();
            prop.enabled = this.literal();
            node.flags.push(prop);
        }
        this.next();
        return node;
    };
    Parser.prototype.property = function () {
        if (this.token.type !== 'name') {
            throw new TypeError("Unexpected token type: " + this.token.type);
        }
        var node = {
            type: 'Property',
            name: this.token.value,
            proptype: '',
            title: EMPTY_LIT,
            deflt: EMPTY_LIT,
            description: EMPTY_LIT,
            choices: [],
        };
        this.next('paren', '(');
        this.next('name');
        node.proptype = this.token.value.toLowerCase();
        this.next('paren', ')');
        if (this.next('name', null, false)) {
            if (['report', 'readonly'].indexOf(this.token.value) !== -1) {
                this.next();
            }
        }
        if (
        // @ts-ignore
        this.token.type === 'symbol' &&
            this.token.value === ':' &&
            this.next('string', null, false)) {
            node.title = this.literal();
        }
        if (
        // @ts-ignore
        this.token.type === 'symbol' &&
            this.token.value === ':' &&
            !this.next('symbol', ':', false)) {
            node.deflt = this.literal();
        }
        if (
        // @ts-ignore
        this.token.type === 'symbol' &&
            this.token.value === ':' &&
            this.next('string', null, false)) {
            node.description = this.literal();
        }
        if (node.proptype === 'choices' &&
            // @ts-ignore
            this.token.type === 'symbol' &&
            this.token.value === '=') {
            this.next('square', '[');
            node.choices = [];
            this.next();
            while (
            // @ts-ignore
            this.token.type !== 'square' ||
                // @ts-ignore
                (this.token.type === 'square' && this.token.value !== ']')) {
                var prop = {
                    type: 'Option',
                    key: this.literal(),
                    value: EMPTY_LIT,
                };
                // @ts-ignore
                if (this.token.type !== 'symbol' || this.token.value !== ':') {
                    throw new TypeError("Unexpected token: " + this.token.type + " " + this.token.value);
                }
                this.next();
                prop.value = this.literal();
                node.choices.push(prop);
            }
            this.next();
        }
        return node;
    };
    Parser.prototype.param = function () {
        if (this.token.type !== 'name') {
            throw new TypeError("Unexpected token type: " + this.token.type);
        }
        var node = {
            type: 'Parameter',
            name: this.token.value,
            properties: [],
        };
        if (this.next('paren', '(', false)) {
            this.next();
            // @ts-ignore
            while (this.token.type !== 'paren' || this.token.value !== ')') {
                switch (this.token.type) {
                    case 'name':
                        node.properties.push(this.name());
                        break;
                    // @ts-ignore
                    case 'string':
                    // @ts-ignore
                    case 'number':
                        node.properties.push(this.literal());
                        break;
                    default:
                        throw new TypeError("Unexpected token type: " + this.token.type + "(" + this.token.value + ")");
                }
                // @ts-ignore
                if (this.token.type === 'symbol' && this.token.value === ',') {
                    this.next();
                }
            }
            this.next();
        }
        return node;
    };
    Parser.prototype.name = function () {
        var node = {
            type: 'Name',
            value: this.token.value,
        };
        this.next();
        return node;
    };
    Parser.prototype.literal = function () {
        var node = {
            type: 'Literal',
            value: this.token.value,
        };
        this.next();
        while (this.token.type === 'symbol' && this.token.value === '+') {
            if (this.next('string', null, false)) {
                node.value += this.token.value;
                this.next();
            }
        }
        return node;
    };
    Parser.prototype.file = function () {
        if (this.token.type !== 'symbol' || this.token.value !== '@') {
            throw new TypeError("Unexpected token type: " + this.token.type);
        }
        this.next('name');
        // @ts-ignore
        if (this.token.value === 'include') {
            return this.include();
        }
        // @ts-ignore
        if (this.token.value === 'mapsize') {
            return this.mapsize();
        }
        // @ts-ignore
        if (this.token.value === 'MaterialExclusion') {
            return this.materialExclusion();
        }
        // @ts-ignore
        if (this.token.value === 'AutoVisGroup') {
            return this.autoVisGroup();
        }
        return this.decl();
    };
    Parser.prototype.parse = function () {
        var ast = {
            type: 'File',
            body: [],
        };
        while (this.current < this.tokens.length) {
            ast.body.push(this.file());
        }
        return ast;
    };
    return Parser;
}());
function parser(tokens) {
    return new Parser(tokens).parse();
}
exports.default = parser;
